#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <stdlib.h>

#include <TROOT.h>
#include <TString.h>
#include <TFile.h>
#include <TF1.h>
#include <TMath.h>

#include <RooGlobalFunc.h>
#include <RooMsgService.h>
#include <RooRealVar.h>
#include <RooCategory.h>
#include <RooFitResult.h>
#include <TMatrixDSym.h>
#include <RooCBShape.h>
#include <RooGenericPdf.h>
#include <RooAddPdf.h>
#include <RooSimultaneous.h>
#include <RooFormulaVar.h>
#include <RooDataSet.h>
#include <RooPlot.h>
#include <RooRandom.h>
#include <RooExtendPdf.h>
#include <RooExponential.h>
#include <RooHist.h>
#include <RooPlotable.h>
#include <RooFitResult.h>
#include <RooMinimizer.h>
#include <RooTrace.h>

using std::cout;
using std::endl;
using std::ends;
using std::fstream;
using std::ifstream; 
using namespace RooFit; 

// Use copy constructor
#define ARG(a) (argList[a])

void fitMttSteph(int massZprime) {

  // Set RooFit verbosity
  RooMsgService::instance().setStreamStatus(0,false);
  RooMsgService::instance().setStreamStatus(1,false);
  RooMsgService::instance().addStream(RooFit::INFO); // DEBUG  INFO  PROGRESS  WARNING ERROR FATAL
  //RooMsgService::instance().setSilentMode(kTRUE);

  //fit region
  Float_t minmTT = 500;
  Float_t maxmTT = 2045;

  RooRealVar Mtt_KF_reco("Mtt_KF_reco", "Mtt_KF_reco", minmTT, maxmTT, "GeV/c^2");
  RooCategory whichLepton("whichLepton", "whichLepton");
  whichLepton.defineType("electron", 11) ;
  whichLepton.defineType("muon", 13) ;

  Double_t lumi_mu = 4678.; 
  Double_t lumi_e = 4682.;
  Double_t b_tagging_correction = 0.90;
  Double_t trigger_correction_muons = 1.;
  Double_t trigger_correction_ele = 1.;
  //
  Double_t muID_correction=0.997; 
  Double_t muIso_correction=0.998; 
  //
  Double_t eleID_correction=1.002; 
  Double_t eleIso_correction=0.9999; 

  double sel_eff_mu = 0.02749410085380077; 
  double sel_eff_e = 0.02008949965238571;
  double hlt_eff_mu = 0.8561673760414124;
  double hlt_eff_e = 0.9664492607116699;
  double s_sel_eff_mu = 0.01431689690798521;
  double s_sel_eff_e = 0.01959382928907871;
  double s_hlt_eff_mu = 0.01051196325841394;
  double s_hlt_eff_e = 0.004138861875743478;

  TFile* f = new TFile("nominal-Zprime750_fitCB.root");

  RooFitResult* fitResult = (RooFitResult*) f->Get("fitResult");
  RooArgList theArgList = fitResult->floatParsFinal();
  TMatrixDSym mat = fitResult->covarianceMatrix();

  // Set all parameters as constant
  RooLinkedListIter iter = theArgList.iterator();
  RooRealVar* var = NULL;
  std::map<std::string, RooRealVar> argList;
  while ((var = (RooRealVar* ) iter.Next())) {
    var->setConstant(true);
    argList.insert(std::pair<std::string, RooRealVar>(var->GetName(), *var));
  }
  
  delete fitResult;

  RooFormulaVar mZ_e("mZ_e", "Zprime mass", "mZ_mu*massShiftEMu", RooArgList(ARG("mZ_mu"), ARG("massShiftEMu")));

  f->Close();
  delete f;

  double eff_mu = trigger_correction_muons * muID_correction  * muIso_correction  * b_tagging_correction * sel_eff_mu * hlt_eff_mu;
  double eff_e  = trigger_correction_ele   * eleID_correction * eleIso_correction * b_tagging_correction * sel_eff_e  * hlt_eff_e;

  RooRealVar lumiRatio("lumiRatio","luminosity ratio",lumi_e/lumi_mu,"");
  RooRealVar effRatio("effRatio","efficiency ratio",eff_e/eff_mu,"");

  RooCBShape sigPdf_mu("sigPdf_mu", "Zprime Crystal Ball", Mtt_KF_reco, ARG("mZ_mu"), ARG("sigma_MZ_mu"), ARG("alphaCB_mu"), ARG("nCB_mu"));
  RooCBShape sigPdf_e("sigPdf_e", "Zprime Crystal Ball", Mtt_KF_reco, mZ_e, ARG("sigma_MZ_e"), ARG("alphaCB_e"), ARG("nCB_e"));

  // configuration of background parametrisation
  // now configured with bkgfit_str parameter
  RooRealVar bkg_param1_mu("param1_mu", "param1_mu",0); bkg_param1_mu.setConstant(kFALSE);
  RooRealVar bkg_param2_mu("param2_mu", "param2_mu",0); bkg_param2_mu.setConstant(kFALSE);
  RooRealVar bkg_param3_mu("param3_mu", "param3_mu",0); bkg_param3_mu.setConstant(kFALSE);
  RooRealVar bkg_param1_e("param1_e", "param1_e",0); bkg_param1_e.setConstant(kFALSE);
  RooRealVar bkg_param2_e("param2_e", "param2_e",0); bkg_param2_e.setConstant(kFALSE);
  RooRealVar bkg_param3_e("param3_e", "param3_e",0); bkg_param3_e.setConstant(kFALSE);
  RooAbsPdf* bkgPdf_mu = NULL; 
  RooAbsPdf* bkgPdf_e = NULL; 
  TString formula;

    bkg_param1_mu.setVal(80.) ; bkg_param1_mu.setRange(-50, 2000.);
    bkg_param2_mu.setVal(-4.) ; bkg_param2_mu.setRange(-50, 2000.);
    bkg_param3_mu.setVal(1.)  ; bkg_param3_mu.setRange(-50, 2000.);
    bkg_param1_e.setVal(80.)  ; bkg_param1_e.setRange(-50, 2000.);
    bkg_param2_e.setVal(-4.)  ; bkg_param2_e.setRange(-50, 2000.);
    bkg_param3_e.setVal(1.)   ; bkg_param3_e.setRange(-50, 2000.);
    formula = TString("pow((1.-(@0/7000)+@3*pow((@0/7000),2)),@1)/pow(@0,@2)");
    bkgPdf_mu = new RooGenericPdf("faltB_mu","faltB_mu", formula.Data(), RooArgList(Mtt_KF_reco, bkg_param1_mu, bkg_param2_mu, bkg_param3_mu));
    bkgPdf_e = new RooGenericPdf("faltB_e","faltB_e", formula.Data(), RooArgList(Mtt_KF_reco, bkg_param1_e, bkg_param2_e, bkg_param3_e));

  // Mtt_KF_reco global PDFs
  RooRealVar nSig_mu("nSig_mu", "number of sig events", 0.,-1000.,2000.); 
  // RooRealVar nSig_mu("nSig_mu", "number of sig events", 0.); //fix signal to zero
  RooFormulaVar nSig_e("nSig_e", "number of sig events", "nSig_mu*effRatio*lumiRatio", RooArgList(nSig_mu,effRatio,lumiRatio));
  // RooRealVar nSig_e("nSig_e", "number of sig events", 0.);
  RooRealVar nBkg_mu("nBkg_mu", "number of bkg events", 2000., 0., 50000);
  RooRealVar nBkg_e("nBkg_e", "number of bkg events", 2000., 0., 50000);

  RooAddPdf globalPdf_mu("global_PDF_mu", "sigPdf+BkgPdf",RooArgList(sigPdf_mu,*bkgPdf_mu), RooArgList(nSig_mu,nBkg_mu));
  RooAddPdf globalPdf_e ("global_PDF_e" , "sigPdf+BkgPdf",RooArgList(sigPdf_e, *bkgPdf_e) , RooArgList(nSig_e ,nBkg_e));

  RooSimultaneous simPdf("simPdf","simultaneous pdf",whichLepton) ;

  // -- pour fitter avec signal+fond 
  simPdf.addPdf(globalPdf_mu,"muon");
  simPdf.addPdf(globalPdf_e,"electron");
  // -- pour fitter avec le fond seulement
  //   simPdf.addPdf(*bkgPdf_mu,"muon");
  //   simPdf.addPdf(*bkgPdf_e,"electron");
  // -- pour fitter avec le signal seulement
  //  simPdf.addPdf(sigPdf_mu,"muon");
  //  simPdf.addPdf(sigPdf_e,"electron");wor
  
  // Get dataset from external file
  TString datafilename = "ds_data_2011-nominal.txt"; 
  RooDataSet *dataOrig= RooDataSet::read(datafilename.Data(), RooArgList(Mtt_KF_reco, whichLepton));

  // Reduce data set
  TString cutstr = TString::Format("Mtt_KF_reco>%d&&Mtt_KF_reco<%d",(int) minmTT,(int) maxmTT);
  RooDataSet* RedData = (RooDataSet*) dataOrig->reduce(cutstr.Data());

  delete dataOrig;

  Double_t minmTTFit = minmTT + 0.0;
  Double_t maxmTTFit = maxmTT - 0.0;
  Mtt_KF_reco.setRange(minmTTFit, maxmTTFit);
  
  // for (int i = 0; i < 500; i++) {

   RooAbsReal* nll = simPdf.createNLL(*RedData, Optimize(0));
   RooMinimizer m(*nll);
   m.setMinimizerType("Minuit");

   // LEAK
   m.optimizeConst(1);

   // NO LEAK
   m.optimizeConst(0);

   //m.setEvalErrorWall(1);
   //m.minimize("Minuit", "migrad");
   //m.cleanup();
   //m.hesse();
   // m.minos();
   delete nll;

   // OR
   // simPdf.fitTo(*RedData, Optimize(0));

//  }

  delete RedData;
  delete bkgPdf_mu;
  delete bkgPdf_e;
}

void memory_leak() {
    RooTrace::active(1);
    fitMttSteph(750);
    RooTrace::dump();
}
